


'''

'''

def IS_AFTER_END (f, LAST_BYTE_INDEX):
	try:
		f.seek (LAST_BYTE_INDEX)
		print ("FOUND LAST BYTE INDEX")		
	except Exception as E:
		print ("Exception1:", E)
		return True
		
	return False

def IS_BEFORE_END (f, LAST_BYTE_INDEX):
	try:
		f.seek (LAST_BYTE_INDEX)
		PLATE = f.read (2)
		if (len (PLATE) == 0):
			return False;
			
		print (f"COULD READ '{ len (PLATE) }' BYTES")
		
	except Exception as E:
		print ("Exception2:", E)
		pass;
				
	return True


def DEEM (PARAMS):
	print ("DEEM IF IS LAST BYTE")

	DRIVE_PATH = PARAMS ["drive path"]
	LAST_BYTE_INDEX = PARAMS ["last byte index"]

	with open (DRIVE_PATH, "rb") as f:
		if (IS_AFTER_END (f, LAST_BYTE_INDEX)):
			return [ False, "THE LAST BYTE DECLARED IS AFTER THE LAST BYTE" ]
		
		f.close ()
		
	print ("-> ISN'T AFTER END")
	
	with open (DRIVE_PATH, "rb") as f:
		if (IS_BEFORE_END (f, LAST_BYTE_INDEX)):
			return [ False, "THE LAST BYTE DECLARED IS BEFORE THE LAST BYTE" ]
			
		f.close ()
		
	print ("-> ISN'T BEFORE END")

	return [ True, "" ]




'''
from BRUNCH.SCULPT.ROUTINES.ROUTINE_2 import ROUTINE_2
'''

def ROUTINE_2 ():
	BYTES_PER_PLATE = 512 * 512;

	BYTES = b''

	LOOP = 1
	while (LOOP <= (BYTES_PER_PLATE / 8)):
		BYTES += b'\xFF\xEF\xFF\xDF\xFF\xCF\xFF\xBF'
		LOOP += 1
		
	return BYTES;




import json

def ERASE_OPTIONS (GROUP):
	import click
	@GROUP.group ("ERASE")
	def GROUP ():
		pass
		
	'''
		BRUNCH ERASE ENTIRELY --drive-path /dev/sde --last-byte-index 40018599423
		BRUNCH ERASE ENTIRELY --drive-path /dev/sde --last-byte-index 32010928127
	'''
	import click
	@GROUP.command ("ENTIRELY")
	@click.option ('--drive-path', required = True, help = '')
	@click.option ('--last-byte-index', required = True, help = '')
	@click.option ('--skip-over', default = 0, help = '')
	def ERASE_ENTIRELY (drive_path, last_byte_index, skip_over):	
		from BRUNCH.ERASE import ERASE
		from BRUNCH.SCULPT.ROUTINES.ROUTINE_1 import ROUTINE_1
		from BRUNCH.SCULPT.ROUTINES.ROUTINE_2 import ROUTINE_2
				
		ERASE ({
			"SKIP OVER": int (skip_over),
			
			"DRIVE PATH": drive_path,
			"LAST BYTE INDEX": int (last_byte_index),
			
			"SCULPT": {
				"ROUTINES": [{
					"BYTES PER PLATE": 512 * 512,
					"BYTES": ROUTINE_1 ()
				},{
					"BYTES PER PLATE": 512 * 512,
					"BYTES": ROUTINE_2 ()
				}]
			}
		})
		
		
		return;

	return;








'''
	import BRUNCH.SCAN as SCAN
'''

'''
	PARAMS = {
		"PLATE": PLATE,
		"BYTES INDEXES": [
			PLATE_INDEX_START,
			PLATE_INDEX_START + SCAN_SIZE
		]
	}
'''
def PROGRESS (PARAMS):
	return;


#
#	https://python-jsonschema.readthedocs.io/en/latest/
#
from jsonschema import validate
SCHEMA = {
    "type" : "object",
	"required": [ 
		"DRIVE PATH",
		"BYTES INDEXES", 
		"BYTES PER PLATE" 
	],
    "properties" : {
		"DRIVE PATH": {
			"type": "string"
		},
        "BYTES INDEXES" : { 
			"type" : "array",
		},
        "BYTES PER PLATE" : {
			"type" : "number"
		}
    }
}

def START (CARGO):
	#print ("CARGO", CARGO)

	validate (
		instance = CARGO, 
		schema = SCHEMA
	)

	DRIVE_PATH = CARGO ["DRIVE PATH"]
	BYTES_PER_PLATE = CARGO ["BYTES PER PLATE"]
	BYTES_INDEXES = CARGO ["BYTES INDEXES"]
	
	#print ("VALID?")
	
	
	if ("PROGRESS" in CARGO):
		PROGRESS = CARGO ["PROGRESS"]
	
	
	#
	#	--
	#

	'''
		BYTES PER PLATE: 10
		BYTE INDEXES: [ 0, 28 ]
	
		 0 TO  9
		10 TO 19
		
		20 TO 29
	'''

	MEAL_INDEX_START = BYTES_INDEXES [0];
	MEAL_INDEX_END = BYTES_INDEXES [1];

	with open (DRIVE_PATH, "rb") as f:
		f.seek (MEAL_INDEX_START)
		#PLATE_INDEX_0 = f.tell ()
		
		print (f"OPENED DRIVE '{ DRIVE_PATH }' FOR SCANNING")
		
		LAST_SCAN = False
		PLATE = True;
		while PLATE:
			PLATE_INDEX_START = f.tell ()
					
			#print (PLATE_INDEX_START, BYTES_PER_PLATE, MEAL_INDEX_END)
					
			#
			#	CHECK IF A FULL PLATE WOULD 
			#	PUT THE SCANNER PAST THE LAST INDEX
			#
			if (
				(PLATE_INDEX_START + BYTES_PER_PLATE) > MEAL_INDEX_END
			):
				LAST_SCAN = True
				SCAN_SIZE = MEAL_INDEX_END - PLATE_INDEX_START + 1

				#print ("PAST LAST INDEX, SCAN_SIZE =", SCAN_SIZE)

				if (SCAN_SIZE <= 0):
					return;
				
			else:
				SCAN_SIZE = BYTES_PER_PLATE;
				
			#print ("SCAN SIZE:", SCAN_SIZE)
			
			PLATE = f.read (SCAN_SIZE)
			if (len (PLATE) == 0):
				print ("THERE WERE NO BYTES LEFT TO SCAN")
				return;	
			
			#print ("TELL:", f.tell ())
			
			PLATE_INDEX_END = f.tell () - 1
			
			PROGRESS ({
				"PLATE": PLATE,
				"SCAN SIZE": SCAN_SIZE,
				"LAST SCAN": LAST_SCAN,
				"INDEXES": [
					PLATE_INDEX_START,
					PLATE_INDEX_END
				]
			})	
			
			if (LAST_SCAN):
				print ()
				print ("SCAN IS DONE")
				print ()
				
				f.close ()
				
				return;
			
			
	
			 
		 







def SCULPT_OPTIONS (GROUP):
	import click
	@GROUP.group ("SCULPT")
	def GROUP ():
		pass
		
	import click
	@GROUP.command ("ENTIRE-DRIVE")
	@click.option ('--drive-path', required = True, help = '')
	@click.option ('--drive-bytes', required = True, help = '')
	def EXAMPLE (port):	
		


		return;
		
		
	'''
		BRUNCH SCULPT BYTES --drive-path /dev/sde --index-0 31914983423 --index-1 31914983423 --hex-string '04'
	'''
	import click
	@GROUP.command ("BYTES")
	@click.option ('--drive-path', required = True, help = '')
	@click.option ('--index-0', required = True, help = '')
	@click.option ('--index-1', required = True, help = '')
	@click.option ('--hex-string', required = True, help = '')
	def SCULPT_BYTES (drive_path, index_0, index_1, hex_string):	
		import BRUNCH.SCULPT as SCULPT

		BYTES_FOR_PLATE = bytes.fromhex (hex_string)
		
		print (BYTES_FOR_PLATE)
		print (type (BYTES_FOR_PLATE))
		

		SCULPT.START ({
			"DRIVE PATH": drive_path,
			
			"BYTES INDEXES": [ 
				int (index_0), 
				int (index_1)
			],		
			"BYTES FOR PLATE": BYTES_FOR_PLATE,
			
			"PROGRESS": lambda PARAMS : (print (PARAMS))
		})

		return;

	return;





'''
from BRUNCH.SCULPT.ROUTINES.ROUTINES.1 import ROUTINE_1
'''

def ROUTINE_1 ():
	BYTES_PER_PLATE = 512 * 512;

	BYTES = b''

	LOOP = 1
	while (LOOP <= (BYTES_PER_PLATE / 8)):
		BYTES += b'\x00\x01\x00\x02\x00\x03\x00\x04'
		LOOP += 1
		
	return BYTES;



'''
	VENTURES:
		
		from BRUNCH.ERASE import ERASE
		from BRUNCH.SCULPT.ROUTINES.ROUTINE_1 import ROUTINE_1
		from BRUNCH.SCULPT.ROUTINES.ROUTINE_2 import ROUTINE_2
		
		ERASE ({
			"SKIP OVER": 0,
			
			"DRIVE PATH": DRIVE_PATH,
			"LAST BYTE INDEX": LAST_BYTE_INDEX,
			
			"SCULPT": {
				"ROUTINES": [{
					"BYTES PER PLATE": 512 * 512,
					"BYTES": ROUTINE_1 ()
				},{
					"BYTES PER PLATE": 512 * 512,
					"BYTES": ROUTINE_2 ()
				}]
			},
			"SCAN": {
				"BYTES PER PLATE": 512 * 512
			}
		})
'''

import BRUNCH.SCAN as SCAN
import BRUNCH.SCAN.IS_LAST_BYTE as IS_LAST_BYTE

import BRUNCH.ERASE.FUNCTIONS.BYTE_STRING_EQ as BYTE_STRING_EQ

import BRUNCH.SCULPT as SCULPT

from fractions import Fraction

def ERASE (CARGO):
	'''
		SKIP_OVER = 0	-> SKIPS NOTHING
		SKIP_OVER = 1 -> SKIPS FIRST SCULPT
		SKIP_OVER = 2 -> SKIPS TO SECOND LOOP
	'''	

	SKIP_OVER = CARGO ["SKIP OVER"]

	DRIVE_PATH = CARGO ["DRIVE PATH"]
	LAST_BYTE_INDEX = CARGO ["LAST BYTE INDEX"]
	SCULPT_ROUTINES = CARGO ["SCULPT"]["ROUTINES"]

	#
	#
	SCULPT_START = 0
	#
	#

	[ IS_LAST, NOTE ] = IS_LAST_BYTE.DEEM ({
		"drive path": DRIVE_PATH,
		"last byte index": LAST_BYTE_INDEX + 1
	})
	if (IS_LAST == False):
		print (NOTE)
	
	print ("LAST_BYTE_INDEX IS_LAST:", IS_LAST)
	


	SCULPT_LOOP = 1
	def SCULPT_PROGRESS (PARAMS):
		nonlocal SCULPT_LOOP;
		INDEXES = PARAMS ["INDEXES"]
		SIZE = PARAMS ["SIZE"]
		LAST_LOOP = PARAMS ["LAST_LOOP"]

		if (SCULPT_LOOP == 1000 or LAST_LOOP == True):
			PERCENT = 100 * Fraction (INDEXES [1] / LAST_BYTE_INDEX)			
			if (PERCENT.denominator == 1):
				PERCENT = str (int (PERCENT))
			else:
				PERCENT = str (float (PERCENT))
			
			
			#BYTE_INDEX = "[AT BYTE INDEX:" + str (INDEXES[0]) + "]"
			PLATE_SIZE = "[REACHED BYTE INDEX:" + str(INDEXES[0]) + "]"
			PERCENT_STRING = "[" + PERCENT + "%]"
			
			print (
				"SCULPTING:",
				PLATE_SIZE,
				PERCENT_STRING
			)
			
			SCULPT_LOOP = 1;
		
		SCULPT_LOOP += 1

	STAGE = 1;
	LOOP_COUNT = 1;
	for ROUTINE in SCULPT_ROUTINES:
		print (f"LOOP '{ LOOP_COUNT }' OF '{ len (SCULPT_ROUTINES) }'")
	
		BYTES_PER_PLATE = ROUTINE ["BYTES PER PLATE"]
		BYTES = ROUTINE ["BYTES"]
		
		#
		#	SKIP_OVER = 0, STAGE = 1	->	DON'T SKIP
		#	SKIP_OVER = 1, STAGE = 1	->	SKIP
		#	SKIP_OVER = 2, STAGE = 1	->  SKIP
		#
		#	SKIP_OVER = 2, STAGE = 3	-> 	DON'T SKIP
		#
		if (STAGE > SKIP_OVER):
			SCULPT.START ({
				"DRIVE PATH": DRIVE_PATH,
				
				"BYTES INDEXES": [ 0, LAST_BYTE_INDEX ],			
				"BYTES FOR PLATE": BYTES,
				
				"PROGRESS": SCULPT_PROGRESS
			})
		else:
			print ("SKIPPING SCULPT, STAGE = ", STAGE)
			
		STAGE += 1;
		
		if (STAGE > SKIP_OVER):
			SCAN_LOOP = 1
			def SCAN_PROGRESS (PARAMS):
				nonlocal SCAN_LOOP;
			
				PLATE = PARAMS ["PLATE"]
				SCAN_SIZE = PARAMS ["SCAN SIZE"]
				INDEXES = PARAMS ["INDEXES"]
				LAST_SCAN = PARAMS ["LAST SCAN"]
				
				# BYTE_STRING_EQ
				
				if (PLATE != BYTES [0:SCAN_SIZE]):
					BYTES_PARTIAL = BYTES [0:SCAN_SIZE]
				
					print ("PLATE LEN:", len (PLATE))
					print ("BYTES LEN:", len (BYTES_PARTIAL))
					
					if (len (PLATE) != len (BYTES_PARTIAL)):
						assert (len (PLATE) != len (BYTES_PARTIAL))
					
					
					INDEX = 0;
					LAST_INDEX = len (PLATE) - 1
					while (INDEX <= LAST_INDEX):
						if (PLATE [ INDEX ] != BYTES_PARTIAL [INDEX]):
							print ("INEQUALITY AT INDEX:", INDEX, PLATE[INDEX], BYTES_PARTIAL[INDEX])
							
						INDEX += 1
					
					assert (PLATE == BYTES [0:SCAN_SIZE])
				
				
				if (SCAN_LOOP % 1000 == 0 or LAST_SCAN == True):
					PERCENT = 100 * Fraction (INDEXES [1] / LAST_BYTE_INDEX)
					if (PERCENT.denominator == 1):
						PERCENT = str (int (PERCENT))
					else:
						PERCENT = str (float (PERCENT))
					
					BYTE_INDEX = "[REACHED BYTE INDEX:" + str (INDEXES[0]) + "]"
					PERCENT_STRING = "[" + str (PERCENT) + "%]"
								
					SIZE = "[SCAN SIZE: " + str (SCAN_SIZE) + "]"			
								
					CONTENT = "[FIRST 16 BYTES: " + PLATE[0:16].hex () + "]"
								
					print (
						"SCANNING:",
						BYTE_INDEX,
						CONTENT,
						PERCENT_STRING
					)
					
				SCAN_LOOP += 1
				
			SCAN.START ({
				"DRIVE PATH": DRIVE_PATH, 
				
				"BYTES INDEXES":  [ 0, LAST_BYTE_INDEX ],	
				"BYTES PER PLATE": BYTES_PER_PLATE,
				
				"PROGRESS": SCAN_PROGRESS
			})
		else:
			print ("SKIPPING SCAN, STAGE = ", STAGE)
			
		STAGE += 1

		LOOP_COUNT += 1
		


	return




#
#	https://setuptools.pypa.io/en/latest/userguide/quickstart.html
#
#	https://github.com/pypa/sampleproject/blob/db5806e0a3204034c51b1c00dde7d5eb3fa2532e/setup.py
#
from setuptools import setup, find_packages

VERSION = "0.0.13"
NAME = 'BRUNCH'
INSTALL_REQUIRES = [
	'jsonschema'
]
	# textual

def SCAN_DESCRIPTION ():
	DESCRIPTION = ''
	try:
		with open ('MODULE.HTML') as f:
			DESCRIPTION = f.read ()
		#print (DESCRIPTION)
	except Exception as E:
		pass;
		
	return DESCRIPTION;


setup (
    version = VERSION,

	name = NAME,
    install_requires = INSTALL_REQUIRES,	
	
	package_dir = { 
		'BRUNCH': 'GALAXY/MODULES/BRUNCH'
	},
	scripts = [ 'GALAXY/MOVES/BRUNCH' ],
	
	license = "",
	long_description = SCAN_DESCRIPTION (),
	long_description_content_type = "text/plain",
)
